/*  This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License , or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

    You may contact the author by:
       e-mail:  hlub@knoware.nl
*/


#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <ctype.h>

#include <string.h>
#include <errno.h>   
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>

/* If you don't have getopt.h, get gnugetopt */

#include <getopt.h>


/* Info about MIDI: http://www.borg.com/~jglatt/tech/midispec.htm */
/* Programming /dev/midi..: http://opensound.com/pguide/ */
#define NOTE_OFF    0x80
#define NOTE_ON     0x90
#define KEYPRESSURE 0xa0
#define CTRLCHANGE  0xb0
#define PROGCHANGE  0xc0
#define AFTERTOUCH  0xd0
#define PITCHWHEEL  0xe0
#define SYSTEM      0xf0

#define EOX         0xf7  /* end of SYSEX message */ 

#define HOLD_PEDAL  64

#define TRUE 	    1
#define FALSE       0

#define NOTE        1
#define COMMAND   (-1)
#define OFF         0



#define HAVE_SNPRINTF
#define BUFFSIZE 512



char *program_name;
void my_writebyte(unsigned char, int);
unsigned char my_readbyte(int);
void handle_sigalarm(int);
void set_alarm(int);
void myerror(const char *, ...);
char *mybasename (char *);
void *mymalloc(size_t); 
char *mysavestring(const char*);



void usage();
char optstring[] =  "ha:c:d:l:npv:s:";

/* Keep track of sounding notes (should really take channel # into
account):*/ 	
int currently_sounding [256]; 
int autorepeating_key = 0; /* currently autorepeating key */
int autorepeat = 0;
int autorepeat_interval_usec = 0;
int autorepeat_delay_usec = 0;
int autoshift_splitpoint = 0;
int pedal_held = 0;


#define AUTOREPEAT_MIN 2
#define AUTOREPEAT_MAX 30
#define AUTOREPEAT_DELAY_MIN 200
#define AUTOREPEAT_DELAY_DEFAULT 500
#define AUTOREPEAT_DELAY_MAX 1000

unsigned char shift_key = 0;
unsigned char command_channel = 0;
int pedal_shifts = FALSE;  
int writeback = TRUE;

int
main (int argc, char **argv)
{
  int midi_fd;
  char *midi_device;
  unsigned char
    maybe_statusbyte,
    statusbyte,
    midibytes[BUFFSIZE], 
    midibyte,
    *midibufptr,
    status_byte_upper_nybble,
    status_byte_lower_nybble,
    split_velocity = 0,
    velocity;
  int 
    c,
    note_number,
    ndatabytes,
    running_status,
    key_type = OFF;
  
  program_name = mysavestring(mybasename(argv[0])); /* needed by myerror() */

  while (1) {
    c = getopt(argc, argv, optstring);
    if (c == EOF)
      break;
    switch(c) {
    case 'a':
      autorepeat = atoi(optarg);
      if (! autorepeat)
	break;
      if (autorepeat >= AUTOREPEAT_MIN && autorepeat <= AUTOREPEAT_MAX) {
	autorepeat_interval_usec = 1000000 / autorepeat;
      } else {
	myerror("Autorepeat rate must be between %d and %d Hz", AUTOREPEAT_MIN, AUTOREPEAT_MAX);
      }
      break;
    case 'c':
      command_channel =  atoi(optarg);
      break;
    case 'd':
      autorepeat_delay_usec = 1000 * atoi(optarg);
      if (autorepeat_delay_usec < 1000 * AUTOREPEAT_DELAY_MIN ||
	  autorepeat_delay_usec > 1000 * AUTOREPEAT_DELAY_MAX)
	myerror("Autorepeat delay must be between %d and %d msec", AUTOREPEAT_DELAY_MIN, AUTOREPEAT_DELAY_MAX);
      break;
    case 'h': 		
      usage(); /* will call exit() */
    case 'l':
      autoshift_splitpoint = atoi(optarg);
      break;
    case 'n':
      writeback = FALSE;
      break;
    case 'p':
      pedal_shifts = TRUE;
      break;
    case 's':
      shift_key = atoi(optarg);
      break;
    case 'v':
      split_velocity = atoi(optarg);
      break;
    }
  }
  if (optind >= argc) 
    usage();
  if (autorepeat && !autorepeat_delay_usec)
    autorepeat_delay_usec = 1000 * AUTOREPEAT_DELAY_DEFAULT;
  
  midi_device = mysavestring (argv[optind]);
  midi_fd = open(midi_device, (writeback ? O_RDWR : O_RDONLY));
  if (midi_fd < 0) 
    myerror("Could not open %s", midi_device);
  
  signal(SIGALRM, &handle_sigalarm);	


  while(1) {  
    midibufptr = midibytes;
    maybe_statusbyte = my_readbyte(midi_fd);
    if (maybe_statusbyte >= 0xf8)
      continue;      /* ignore realtime messages */
    if (maybe_statusbyte & 0x80)  { 
      statusbyte = maybe_statusbyte;
      running_status = FALSE;
    } else { /* running status, i.e. statusbyte omitted, last status implied */
      running_status = TRUE;
      *midibufptr++ = maybe_statusbyte;	         
    }
    if (! running_status) {
      status_byte_upper_nybble = statusbyte & 0xf0;
      status_byte_lower_nybble = statusbyte & 0x07;
      switch (status_byte_upper_nybble) {
      case NOTE_ON:
      case NOTE_OFF:	
      case KEYPRESSURE:
      case CTRLCHANGE:
      case PITCHWHEEL: ndatabytes = 2;  
	break;
      case PROGCHANGE:
      case AFTERTOUCH: ndatabytes = 1;
	break;
      case SYSTEM: 	 
	switch (status_byte_lower_nybble) {
	case 0: /* sysex, read until EOX and discard */
	  my_writebyte (statusbyte, midi_fd);
	  do {
	    midibyte = my_readbyte(midi_fd);
	    my_writebyte(midibyte, midi_fd);
	  } while (midibyte != EOX);
	  continue;	
	case 2: /* song position, is this ever sent by a MIDI keyboard? probably not */  
	  ndatabytes = 1;
	default:  
	  ndatabytes = 0;
	}
	break;
      }
    }
    for ( /* midibufptr already set */ ; 
	  midibufptr < midibytes + ndatabytes; 
	  midibufptr++) {
      *midibufptr = my_readbyte(midi_fd);      
    }

    if (status_byte_upper_nybble  == CTRLCHANGE &&
	midibytes[0] == HOLD_PEDAL) {
      pedal_held = (midibytes[1] > 63);
    }
    if (status_byte_upper_nybble == NOTE_ON || 
	status_byte_upper_nybble == NOTE_OFF ) { 
      
      note_number = midibytes[0];
      velocity    = midibytes[1];
      if (status_byte_upper_nybble == NOTE_ON && velocity > 0) { /* Real NOTE_ON event */
	key_type = NOTE; 
	if (split_velocity  && velocity > split_velocity)
	  key_type *= -1; /* remember: NOTE == 1 and COMMAND == -1 */
	if (shift_key  && currently_sounding [shift_key] == NOTE)
	  key_type *= -1;
	if (pedal_shifts && pedal_held)
	  key_type *= -1;
	if (autoshift_splitpoint && note_number <= autoshift_splitpoint)
	  key_type *= -1; 
	if (command_channel && status_byte_lower_nybble == (command_channel - 1))
	  key_type *= -1;

	if (key_type == COMMAND)
	  midibytes[1] = 0; /* don't play this note */
	
	currently_sounding[note_number] = key_type;
	if (!(note_number == shift_key && shift_key && key_type == NOTE)) {	/* ignore shift key being pressed */
	  printf("%d\n", note_number * key_type); 
	  if (autorepeat && key_type == COMMAND) {                     		/* only commands ... */
	    autorepeating_key = note_number;
	    set_alarm(autorepeat_delay_usec);   	     			/* ... are autorepeating */
	  }
	}
      } else {	                   	/* Real NOTE_OFF (or pesudo one, i.e. NOTE_ON with velocity 0) */

	if (shift_key &&  		/* upon release of shift key ... */
	    note_number == shift_key &&	/* ... when no command has been seen */
	    (key_type != COMMAND)) { 	/* (key_type remembers last NOTE_ON) */
	  printf("%d\n", note_number);  /* print it better late than never! */
	}
	currently_sounding[note_number] = OFF;	
      }
    } /* endif ... NOTE_ON || NOTE_OFF */

      
    if (! running_status) 
      my_writebyte (statusbyte,midi_fd);

    for ( midibufptr = midibytes; 
	  midibufptr < midibytes + ndatabytes; 
	  midibufptr++) {
      my_writebyte(*midibufptr, midi_fd);	
    }		
  }
}


unsigned char my_readbyte(int fd) {
  unsigned char byte;
  ssize_t nread;
  for (;;) {
    nread = read (fd, &byte, 1);
    if (nread > 0)
      return byte;
    if (nread == 0 || (nread< 0 && (errno  == EINTR)))
      continue;
    else
      myerror("Read error");
  }
}

void my_writebyte (unsigned char byte, int fd) {  
  ssize_t nwritten;
  if (!writeback)
    return;
  for (;;) {
       nwritten = write(fd, &byte, 1);
       if (nwritten > 0)
	 return;
       if (nwritten == 0 || (nwritten < 0 && (errno = EINTR)))
	 continue;
       else
	 myerror ("Write error");
     }
}


void
handle_sigalarm(int signo)  
{
  sigset_t mask;
  int saved_errno, error;

  saved_errno = errno;
  if (autorepeating_key && (currently_sounding[autorepeating_key] == COMMAND)) {
     printf("%d\n", - autorepeating_key);   	
     signal(signo, &handle_sigalarm);	
     set_alarm(autorepeat_interval_usec);
  } else {
    autorepeating_key = 0;
  }
  errno=saved_errno;
}


void
set_alarm(int interval)
{
  struct  itimerval  timer = {{0,0},{0,0}};
  if (! interval)
    return;
  timer.it_value.tv_usec = interval;
  if (setitimer(ITIMER_REAL, &timer, NULL))
    myerror("Couldn't set timer");	

}


void usage() {
  printf(
"Usage: %s OPTION DEVICE\n"
"-a N  Autorepeat rate in repeats/sec; 0 = disable\n"
"-d N  Autorepeat delay in msec\n"
"-c N  All notes on MIDI channel N are shifted (0 = disable)\n"
"-h    This help message\n"
"-l N  All keys #M (M <= N) are shifted\n"
"-n    Don't try to write back midi events\n"
"-p    Use hold pedal to shift notes\n"
"-s N  When MIDI key #N is held, all other notes are shifted\n"
"-v N  Any note louder than velocity N is shifted\n"
"\n"
"DEVICE should be a raw MIDI device (/dev/midi?? on Linux)\n\n"
, program_name);
  exit(0);
}


void
myerror(const char * message, ...)
{
  va_list ap;
  int saved_errno = errno;
  char buffer[BUFFSIZE];
  
  va_start(ap, message);

#ifdef  HAVE_SNPRINTF
  snprintf(buffer, sizeof(buffer) - 1, "%s: ", program_name);
  vsnprintf(buffer + strlen(buffer),sizeof(buffer)-strlen(buffer)-1,
            message, ap);
  if (saved_errno)
    snprintf(buffer + strlen(buffer) ,sizeof(buffer)-strlen(buffer)-1,
             ": %s", strerror(saved_errno));
#else
  sprintf(buffer, "%s: ", program_name);
  vsprintf(buffer + strlen(buffer),
            message, ap);
  if (saved_errno)
    sprintf(buffer + strlen(buffer) ,
             ": %s", strerror(saved_errno));
#endif
  strcat(buffer, "\n");
  
  fflush(stdout);
  fputs(buffer, stderr);
  
  va_end(ap);
  exit (-1);
}

char *
mybasename (char *name)
{ /* determine basename of "name" */

  
  char *p;
  /* find last '/' in name (if any) */
  for (p=name + strlen(name) - 1; p > name; p--)
    if (*(p-1) == '/')
      break;            
  return p;
}

void * 
mymalloc(size_t size) 
{ /* malloc with error handling (i.e. bail out when out of memory) */
  void *ptr;
  ptr = malloc(size);
  if (ptr == NULL) 
    myerror("Tried in vain to allocate %d bytes", size);
  return ptr;
}

char *
mysavestring(const char* string)
{ /* allocate copy of string on the heap */ 
  char *buf;
  buf = (char *) mymalloc (strlen (string) + 1);
  strcpy(buf, string);
  return buf;
}
