;;;
;;; midi-input.el --- Minor mode for typing note names in Emacs 
;;;                   with a MIDI keyboard  
;;; 
;;; (c) 2001 Hans Lub <hlub@knoware.nl>
;;;                   
;;;
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2 of
;;; the License, or (at your option) any later version.

;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the implied
;;; warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
;;; PURPOSE.  See the GNU General Public License for more details.
          
;;; You should have received a copy of the GNU General Public
;;; License along with this program; if not, write to the Free
;;; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
;;; MA 02111-1307 USA
  

;;; Put the following line in your .emacs file
;;; (require 'midi-input)

(provide 'midi-input)

;(setq debug-on-error t)

;; The following macro is used in some of the following defcustom
;; definitions to set a variable which needs some extra action
;; to become effective (restart the MIDI reader, say).
(defmacro set-and-then (some-action)
  `(lambda (var val) 
     (progn 
       (set var val) 
       ,some-action)))



(defgroup midi-input nil 
"Use a MIDI keyboard to type notenames in Emacs."
  :group 'environment
  :group 'multimedia
  :prefix "midi-input-")

;;;###autoload
(defcustom midi-input-mode nil  
  "Toggle midi-input-mode.
Setting this variable directly does not take effect;
use either \\[customize] or the function `midi-input-mode'."
  :set (lambda (symbol value)
	 (message "Value %S" value)
     	 (midi-input-mode (if value 1 0)))
  :initialize 'custom-initialize-default  
  :type    'boolean
  :group   'midi-input
  :require 'midi-input)
           
(defgroup midi-input-shift-methods nil 
"How to use the MIDI keyboard to type strings and give commands"
  :group 'midi-input
  :prefix "midi-input-")

(defcustom midi-input-pedal-shifts nil
  "*Use hold pedal to shift keys" 
  :group 'midi-input-shift-methods
  :type   '(choice (const :tag "No" nil)
		  (const :tag "Yes"  t))  
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-restart)))

(defcustom midi-input-shift-key 36
  "*MIDI key number of shift key (C=36); set to 0 to disable.
While this key is held any other keypress will shifted, i.e. treated
as a command. If you don't press any other key *releasing* the
shift key will enter its note"
  :group 'midi-input-shift-methods
  :type	'(integer :tag "MIDI key number")
	;	  :action midi-input-check-keyno 
        ;	  :format "%[note?%] %v"))	
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-restart)))

(defcustom midi-input-command-channel 0
  "*MIDI channel for shifted keys (ranging 1-16); 0 means disable.
Probably only useful if you have a double keyboard"
  :group 'midi-input-shift-methods
  :type 'integer
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-restart)))

(defcustom midi-input-shift-velocity 0
  "*Any note louder than this is shifted; set to 0 to disable"
  :group 'midi-input-shift-methods
  :type 'integer
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-restart)))


(defcustom midi-input-autorepeat-rate 0
  "*Autorepeat rate (in repeats/sec) for command keys"
  :group 'midi-input-shift-methods
 :type  '(choice (const :tag "no autorepeat" 0)
		  (const :tag "5/sec"  5)
		  (const :tag "10/sec"   10)
		  (const :tag "25/sec" 25))
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-restart)))


(defcustom midi-input-autorepeat-delay 500
  "*Autorepeat delay (in msec, between 100 and 1000) for command keys"
  :group 'midi-input-shift-methods
  :type   '(choice (const :tag "0.1 sec" 100)
		  (const :tag "0.3 sec"  300)
		  (const :tag "0.5 sec"   500)
		  (const :tag "1 sec" 1000))
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-restart)))


(defgroup midi-input-notation nil 
"Notation conventions"
  :group 'midi-input
  :prefix "midi-input-")

(defcustom midi-input-notenames '("bes" "f" "c" "g" "d" "a" "e" "b")
  "*8 notenames along the circle of fifths, upwards from b flat"
  :group 'midi-input-notation
  :type  '(list (string :tag "b flat")
	      (string :tag   "f     ")
	      (string :tag   "c     ")
	      (string :tag   "g     ")
	      (string :tag   "d     ")
	      (string :tag   "a     ") 
	      (string :tag   "e     ")
	      (string :tag   "b     "))
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-initialise)))

(defcustom midi-input-flat-suffix  "es"
  "*Suffix indicating a flat"
  :group 'midi-input-notation
  :type 'string
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-initialise)))


(defcustom midi-input-sharp-suffix "is"
  "*Suffix indicating a sharp"            
  :group 'midi-input-notation
  :type 'string)

(defcustom midi-input-octave-up-suffix "'"
  "*Suffix indicating (a jump to) a higher octave"
  :group 'midi-input-notation
  :type 'string)


(defcustom midi-input-octave-down-suffix ","
  "*Suffix indicating (a jump to) a lower octave"
  :group 'midi-input-notation
  :type 'string)

(defcustom midi-input-relative-mode t
  "*Determine octave-(up|down)-suffix relative to previous note"
  :group 'midi-input-notation
  :type  '(choice (const :tag "Yes" t)
                  (const :tag "No" nil)))

(defcustom midi-input-absolute-mode-c-note "c'"
  "*If not in relative mode, output this note as \"c\""
  :group 'midi-input-notation
  :type  '(choice (const :tag "c''" "c''")
		  (const :tag "c'"  "c'")
		  (const :tag "c"   "c")
		  (const :tag "c,"  "c,"))
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-initialise)))



(defcustom midi-input-device "/dev/midi"
  "*Raw MIDI device"
  :group 'midi-input
  :type 'string
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-restart)))
  
(defcustom  midi-input-reader-program "midi-input"
  "*MIDI input reader program"
  :group 'midi-input
  :type 'string
  :initialize 'custom-initialize-default
  :set (set-and-then (midi-input-restart)))


(defcustom midi-input-shifted-keys
  
  '((37     (delete-backward-char 1) "Delete")    ; Cis 
    (38      " |\n" "barcheck | and newline")     ; D
    (39      " "  "space" )                       ; Es       
    (40      "r" "")                              ; E
    (41     (midi-input-replace-last-note 'midi-input-enharmonise) "Enharmonise last note before point" ) ;F
    (42     (midi-input-find-pattern) "Find pattern in current selection" ) ; Fis
    (43     (midi-input-replace-pattern) "Replace notes before point by current pattern" ) ; G
    (45     (undo) "Undo! (doesn't work right yet. Why??)") ; A
    (47     (midi-input-toggle-help) "Display/remove this help screen") ; B
    
    (48     "1" "")      ; c
    (49     "(" "")      ; cis
    (50     "2" "")      ; d
    (51     ")" "")      ; es
    (52     "4" "")      ; e

    (53     "8" "")      ; f 
    (54     "." "")      ; fis
    (55     "16" "")     ; g
    (56     "[" "")      ; gis
    (57     "32" "")     ; a
    (58     "]" "")      ; bes
    (59     "64" "")     ; b
    
   
    (60     (midi-input-replace-pattern "%s()%s %s()%s") "pattern a()a a()a")
    (62     (midi-input-replace-pattern "%s(%s %s)%s") "pattern a(a a)a")
    (64     (midi-input-replace-pattern "%s(%s)%s") "pattern a(a)a")
    (65     (midi-input-replace-pattern "%s4. %s8 %s4. %s8") "pattern a4. a8 a4.a8")


    (72      (unless (bobp) (backward-char)) "cursor left")
    (73      (newline)       "newline")
    (74      (next-line 1)     "cursor down")
    (75      (previous-line 1) "cursor up")
    (76      (unless (eobp) (forward-char))  "cursor right"))


 
         
"*Text or actions bound to shifted keys"

  :group 'midi-input
  :type '(repeat :tag "Shifted keys"
	  (list :tag "Shifted key"
		(integer :tag "MIDI key number" 
			 :action midi-input-check-keyno 
			 :format "%[note?%] %v")
		(sexp    :tag "String or elisp expression")
		(string  :tag "Documentation"))))

(defun midi-input-check-keyno (widget event)
  (let ((value (widget-value widget)))
    (cond ((or (< value  0) (> value  120))
	   (progn 
	     (message "There is no MIDI note #%d" value)
	     nil))
	  (t 
	   (message "Note #%d is %s" value 
		    (midi-input-noteno-to-notename value t))))))

(defvar midi-input-user-defined-shifted-keys nil)

(defvar midi-input-notename-regexp  nil
  "Regexp matching any notename without octave suffixes.
Set by `midi-input-initialise'")

(defvar midi-input-notename-in-context-regexp  nil
  "Regexp matching any notename in context.

Used to prevent a match e.g. on the \"e\" or \"a\" in \"easy\"
the notename itself must be the 2nd parenthesised subexpression!
Set by `midi-input-initialise'") 

(defvar midi-input-before-notename-regexp  "^\\|[^a-z\\]") ; may not contain parethtesised subexpr!
(defvar midi-input-after-notename-regexp   "$\\|[^a-z\\]")
;(defvar midi-input-absolute-mode-offset)


(defvar midi-input-process nil
  "Process object for`midi-input-reader-program'") 

(defvar midi-input-mode nil
  "Non-nil if MIDI input minor mode is enabled.
Use the command \\[midi-input-mode] to change this variable")

(defvar midi-input-tonal-centre 1600
  "Moving avarage of last few notes' position on the circle.
scaled by a factor 100, i.e. c = 1600

Used to determine the spelling of a new note '(e.g. f sharp or g flat)")

(defvar midi-input-lastnote 60
  "noteno of previous note; determines octave suffixes in relative mode")

(defvar midi-input-lastnum 0
  "Position on the circle of  previous note.
Used to determine octave suffix after diminished fifth or
augmented fourth")

(defvar midi-input-fortissimo 90
  "Notes louder than this velocity are shifted.
Used in command-line of `midi-input-reader-program'" ) 

(defvar midi-input-proc-buffername "MIDI input"
  "Name of process buffer for `midi-input-reader-program'.")

(defvar midi-input-notelist-c-offset nil
  "Key value of c '(bis, ceses) in `midi-input-notelist', modulo 12")		  




(defun midi-input-list-filter (list fun)
  (if list
      (let ((head (car list)))
	(if (funcall fun head)
	    (cons head (midi-input-list-filter (cdr list) fun))
	  (midi-input-list-filter (cdr list) fun)))
    nil))

;;; TRYME (midi-input-list-filter '(0 1 2 3 4 5 6 7 8) '(lambda (x) (< 4 x)))


(defun midi-input-add-shifted-key (key val &optional documentation)
  "Add '\(KEY VAL DOCUMENTATION) to `midi-input-user-defined-shifted-keys',
VAL = nil means: remove KEY"
  (let ((noteno (midi-input-notename-to-noteno key)))
    (if val
	(setq midi-input-user-defined-shifted-keys
	    (cons (if documentation
		      (list noteno val documentation)
		    (list noteno val ""))
		  midi-input-user-defined-shifted-keys))
      (setq midi-input-user-defined-shifted-keys
	    (midi-input-list-filter midi-input-user-defined-shifted-keys
				    '(lambda (x) (not (equal noteno (car x)))))))))


;;; TRYME (midi-input-add-shifted-key "b'" "Arrgh!" "scary...")

(defun midi-input-mode (&optional arg)
  "Enable/disable MIDI input in all buffers"
  (interactive "P")
  (setq midi-input-mode
	(if (null arg) (not midi-input-mode)
	  (> (prefix-numeric-value arg) 0)))
  (add-to-list  'minor-mode-alist  '(midi-input-mode " Midi"))
  (cond ((and midi-input-mode (not midi-input-process))	 
	 (if (not (midi-input-start midi-input-device))
	     (progn
	       (message "Couldn't start reader process")
	       (pop-to-buffer midi-input-proc-buffername)
	       (midi-input-mode nil)))) ; start failed
	((and midi-input-process (not midi-input-mode))
	 (midi-input-stop)))) 

(defun midi-input-reverse-acc (list acc)
  (if list
      (midi-input-reverse-acc (cdr list) (cons (car list) acc))
    acc)) 

(defun midi-input-reverse (list) 
  (midi-input-reverse-acc list nil))

(defun midi-input-flatten (notename)
"Appends \\[midi-input-flat-suffix] to notename, contracting double vowels like ees (-> es)"
  (let ((flatnotename (concat notename midi-input-flat-suffix)))
    (if (string-match "\\([aeiou]\\)[aeiou]" flatnotename) ; drop last of two consecutive vowels 
	(replace-match "\\1" t nil flatnotename)
      flatnotename)))

(defun midi-input-sharpen (notename)
"Appends \\[midi-input-sharp-suffix] to notename"
  (concat notename midi-input-sharp-suffix)) 



(defun midi-input-list-to-assoclist-2 (list n)
  (if list
      (cons  (cons n (car list)) (midi-input-list-to-assoclist-2 (cdr list) (+ n 1)))
    nil))


(defun midi-input-list-to-assoclist (list)  
  (midi-input-list-to-assoclist-2 list 0))

(defun midi-input-replace-last-note (fun)
  "Replace last note <note> before point with (FUN <note>)"
  (catch 'badnote
    (save-excursion
      (let*( (linestart 
	      (save-excursion (progn (beginning-of-line) (point)))) 
	     (regexp 
	      (format "\\(%s\\)\\(%s\\)" midi-input-before-notename-regexp midi-input-notename-regexp))
	     (matchpos
	      (condition-case nil
		  (re-search-backward regexp linestart)
		(error (throw 'badnote (message "Couldn't find note before point")))))	    
	     (note
	      (match-string 2))
	     (newnote 
	      (funcall fun note)))
	(replace-match newnote t t nil 2)))))
    

(defun midi-input-make-notelist () 
  (let* ((upper7
	  (cdr midi-input-notenames )) ; ("f" "c" "g" "d" "a" "e" "b")
	 (lower7
	  (midi-input-reverse
	   (cdr (midi-input-reverse midi-input-notenames)))) ; ("bes" "f" "c" "g" "d" "a" "e")
	 (flats
	  (mapcar 'midi-input-flatten lower7))
	 (double-flats
	  (mapcar 'midi-input-flatten flats))
	 (sharps
	  (mapcar 'midi-input-sharpen upper7))
	 (double-sharps
	  (mapcar 'midi-input-sharpen sharps))
	 (all-of-them
	  (append double-flats flats midi-input-notenames sharps double-sharps)))
    (midi-input-list-to-assoclist all-of-them)))


(defvar midi-input-notelist nil
  "Circle of fifths (unwound) represented as an assoclist")
(defvar midi-input-notelist-length 0)


(defun midi-input-num-to-notename (num)
  "Convert number to notename (along circle of fifths)
   num may be any number, even negative. (We cannot modulate off the circle)"
  (let ((n  (cond ((< num 0)          
		       (mod num 12)) 
		  ((>= num midi-input-notelist-length)
                       (let* ((diff
			       (- num midi-input-notelist-length))
			      (octdown
			       (1+ (/ diff 12)))
			      (notedown
			       (* 12 octdown)))
			 (- num notedown)))
		  (num))))
    (cdr (assoc n  midi-input-notelist))))

(defun midi-input-notename-to-num (name)
  "Convert notename to notenum"
  (car (rassoc name midi-input-notelist))) 

(defun midi-input-transpose (num name)
  "Transpose note name num fifths along circle"
  (midi-input-num-to-notename (+ (midi-input-notename-to-num name) num)))

;; TRYME (midi-input-notename-to-num "ais")
;; TRYME (midi-input-num-to-notename 26)
;; TRYME (midi-input-transpose "d" -2)

(defun  midi-input-distance-along-circle (name1 name2)
  "Find distance along circle of fifths"
  (- (midi-input-notename-to-num name1) (midi-input-notename-to-num name2)))


(defun midi-input-count (el list)
  "count number of elements equal to el in list"
  (let ((tail 
	 (member el list)))
    (if tail
	(1+ (midi-input-count el (cdr tail)))
      0)))
	


(defun midi-input-notename-to-noteno   (name-with-suffix)
  "Convert absolute notename to MIDI note number (e.g. c'' -> 72)"
  (let* ((octave-up-or-down-suffixes 
	  (concat  midi-input-octave-up-suffix midi-input-octave-down-suffix))
	 (name-and-suffix-regexp 
	  (format "\\([^%s]*\\)\\([%s]*\\)" octave-up-or-down-suffixes octave-up-or-down-suffixes ))   
	 (match 
	  (string-match name-and-suffix-regexp name-with-suffix))
	 (name   
	  (match-string 1 name-with-suffix))
	 (suffix 
	  (match-string 2 name-with-suffix))
	 (notenum    
	  (midi-input-notename-to-num name))
	 (noteno-mod-12 
	  (mod (* 7 (- notenum  midi-input-notelist-c-offset)) 12))
	 (modifiers 
	  (split-string suffix ""))
	 (octaves 
	  (- (midi-input-count midi-input-octave-up-suffix  modifiers) 
	     (midi-input-count midi-input-octave-down-suffix modifiers))))
    (+ 48  noteno-mod-12 (* 12 octaves))))

;;; TRYME (midi-input-notename-to-noteno "c''")

(defun midi-input-enharmonise (name &optional turns)
  "Find NAME in circle of fifths, and turn TURNS times full circle
TURNS > 0 means toward sharps, < 0 toward flats, nil means
towards C . Calling this function  will also bump
\\[midi-input-tonal-centre] in the corresponding  direction"

  (let* ((n 
	  (midi-input-notename-to-num name))
	 (c 
	  (midi-input-notename-to-num "c"))
	 (nturns
	  (if turns turns (if (<= n c) +1 -1))))
    (if n
	(progn
	  (setq midi-input-tonal-centre (+ midi-input-tonal-centre (* 800 nturns)))
	  (midi-input-num-to-notename (+ (* 12 nturns) n)))
      name)))




(defun midi-input-process-filter (proc string)
  (let*
      (  (previous 
	  (char-to-string (if (char-before) (char-before) 0)))
	 (new 
	  (midi-input-digest string))
	 (space 
	  (if (and (string-match "[a-hrss0-9\\.',]" previous)
		   (string-match "[a-hrs]" new))
	      " " "")))
	 
    (insert (concat space new))
    (if (> (length new) 0)   (undo-boundary))))  


(defun midi-input-zipwith (fun list1 list2) 
  (if (or (null list1) (null list2))
      nil
    (cons (funcall fun (car list1) (car list2))
	  (midi-input-zipwith fun (cdr list1) (cdr list2)))))

;;; TRYME (midi-input-zipwith '* '(1 2 3 4 5) '(10 20 30 40 50)) 


(defun midi-input-start (device)
  "Start MIDI reader subprocess"
  (message "Starting MIDI input")
  (let*  ( (process-connection-type 
	    t)	   
	   (reader-argstr
	    (format "-s %s -v %s -c %s  -a %d -d %d %s %s"
		    midi-input-shift-key
		    midi-input-shift-velocity
		    midi-input-command-channel
		    midi-input-autorepeat-rate
		    midi-input-autorepeat-delay
		    (if midi-input-pedal-shifts " -p " "")
		    device))
	   (reader-arglist
	    (split-string reader-argstr))
	   (midiproc
	    (condition-case nil
		(apply 'start-process "midiproc" midi-input-proc-buffername
		       midi-input-reader-program reader-arglist)
	    (error nil))))
    (sleep-for 0.5)
    (if (not (and midiproc (equal (process-exit-status midiproc) 0)))
	nil
      
      (set-process-filter midiproc 'midi-input-process-filter)
      (midi-input-initialise)
      (setq midi-input-process midiproc))))
  
(defun midi-input-stop ()
  "Stops midi input by killing reader process"
  (message "Stopping MIDI input")
  (delete-process midi-input-process)
  (setq midi-input-process nil))

(defun midi-input-restart ()
  "If in midi-input mode, switch it off and on again"
  (if midi-input-mode
      (progn
	(midi-input-mode 0)
	(midi-input-mode 1))))

(defun midi-input-initialise ()
  "Initialise the midi-input data structures"
  (interactive)
  (setq midi-input-notelist 
	(midi-input-make-notelist))
  (setq midi-input-notelist-length 
	(length midi-input-notelist))
  (setq midi-input-notename-regexp 
	(regexp-opt (mapcar 'cdr midi-input-notelist)))
  (setq midi-input-notename-in-context-regexp
       (format "\\(%s\\)\\(\\(%s\\)\\(:?%s+\\|%s+\\)?\\)\\(%s\\)" 
	       midi-input-before-notename-regexp 
	       midi-input-notename-regexp 
	       midi-input-octave-up-suffix
	       midi-input-octave-down-suffix
	       midi-input-after-notename-regexp))
  (setq midi-input-lastnum 
	(midi-input-notename-to-num "c"))
  (setq  midi-input-notelist-c-offset 
	 (mod midi-input-lastnum  12)) 
  (setq midi-input-absolute-mode-offset
	(midi-input-notename-to-noteno midi-input-absolute-mode-c-note)))

;; TRYME (midi-input-initialise)

(defun midi-input-noteno-to-notename (noteno &optional absolute)
 "Convert NOTENO (e.g. 66) to string (e.g. \"fis\", or \"ges\",
depending (if ABSOLUTE is non-nil) on previously seen note and
current `midi-input-tonal-centre'"
  
  (if (< noteno 0) nil
    (let*((noteno_mod_12
	   (% (+ midi-input-notelist-c-offset noteno) 12))
	  (tonal-centre
	   (if absolute  1700 midi-input-tonal-centre)) ; 1700 = G (I prefer cis to des)
	  (notenum
	   (midi-input-nearest  (midi-input-divide-modulo  noteno_mod_12 7 12)
				(/ tonal-centre 100) 12))	  
	  (name
	   (midi-input-num-to-notename notenum))    
	  (lastnote 
	   (cond
	    (absolute 54)               ; absolute mode is treated like relative mode where the last
		                        ; note happens to have been a "fis" (= noteno 54) 
	    (midi-input-relative-mode midi-input-lastnote)
	    (t midi-input-absolute-mode-offset)))
	  (note-diff
	   (- noteno lastnote))
	  (direction-along-keyboard
	   (cond ((< note-diff 0) -1) ((> note-diff 0) 1) (t 0)))
	  (num-diff
	   (-  notenum (if (or absolute (not midi-input-relative-mode))
			   (midi-input-notename-to-num "fis")
			 midi-input-lastnum)))
	  (direction-along-circle 
	   (cond ((< num-diff 0) -1) ((> num-diff 0) 1) (t 0)))
	  (octave-diff
	   (/ 
	    (+ (* 2 note-diff) (* 12 direction-along-keyboard)  (- direction-along-circle))
	    24)) 
	  (suffix
	   ( if (< octave-diff 0)
	       (apply 'concat (make-list (- octave-diff) midi-input-octave-down-suffix))
	     (apply 'concat (make-list octave-diff midi-input-octave-up-suffix)))))
      
      (if (not absolute)
	  (progn
	   (setq midi-input-tonal-centre 
		 (/ 
		  (+ (* 100 notenum) 
		     (* 2 midi-input-tonal-centre)) 
		  3))
	   (setq midi-input-lastnote noteno)
	   (setq midi-input-lastnum notenum)))
      (concat name suffix))))

;;; TRYME (midi-input-noteno-to-notename 65)  

	
(defun midi-input-digest (string)
"Translate output from reader process into string or action.
STRING should be a string starting with a number" 
  (let* ( (noteno-maybe-negative
	   (string-to-number string))			    	  
	  (fullname
	   (midi-input-noteno-to-notename noteno-maybe-negative))	   
	  (lookup-first
	   (assoc (- noteno-maybe-negative) midi-input-shifted-keys)) ; Undocumented feature (ahem:)
	  (lookup                                                     ; if you put a *negative* number in the 
					                              ; shifted-key list, the *ordinary* key will 
	  (if lookup-first                                            ; trigger the action. (not very useful)
	       lookup-first
	     (assoc (- noteno-maybe-negative)
		    midi-input-user-defined-shifted-keys))))
    (cond (fullname 
	   fullname)
	  (lookup 
	   (let( (value (cdr lookup)))
	     (cond ((stringp value)         value)                         
		   ((stringp (car value) ) (car value))                   
		   ( t                     (progn (eval (car value))"")))))
	  (t ""))))
		           


;;; TRYME (midi-input-digest "49")                                          

(defun midi-input-divide-modulo (dividend divisor modulus)
"Returns DIVIDEND/DIVISOR in Z/modulus, nil if undefined (e.g. 1/2 in Z/4)"
  (let* ((acc 0)
	(count 0))
   (while (and (/= dividend (mod acc modulus)) (< count modulus))
     (setq acc (+ divisor acc))
     (setq count (1+ count)))
   (if (< count modulus) count nil)))

;;; TRYME (midi-input-divide-modulo 6 7 12)

  
(defun midi-input-nearest (num centre modulus)
"Returns the number nearest to CENTRE which is equal to NUM mod MODULUS"
  (let* (  (x
	    (+ num (- centre (mod centre modulus))))
	   (lower
	    (if (< x centre) x (- x modulus)))
           (diff
	    (- centre lower)))
         (if (< diff (/ modulus 2)) lower (+ modulus lower))))

;;; TRYME (midi-input-nearest 3 100 10)

(defun midi-input-splice (list separator)
"Splice SEPARATOR between the members of LIST.
E.g. (midi-input-splice '(1 2 3 4) 0) = (1 0 2 0 3 0 4)"
  (cond ((null list) nil) 
	((equal 1 (length list)) list)
	(t (cons
	    (car list)
	    (cons
	     separator
	     (midi-input-splice (cdr list) separator))))))

;;; TRYME  (midi-input-splice '(1 2 3 4) 0)

(defvar midi-input-current-replacement-pattern "%s") ; "%s" is a no-op

(defun midi-input-find-pattern ()
  (if mark-active
      (midi-input-find-pattern-command (region-beginning) (region-end))
    (message "Please select a pattern first!")))
  

(defun midi-input-find-pattern-command (beg end)
"Extract a pattern from the region and make it the current replacement pattern"
  (interactive "r")
  (let*( (region
	  (buffer-substring-no-properties beg end))
	 (nmatches-and-pattern
	   (midi-input-replace-all-matches midi-input-notename-in-context-regexp region (lambda (str) "%s") 2))
	 (pattern
	  (car nmatches-and-pattern))
	 (nmatches
	  (cdr nmatches-and-pattern)))
    (setq  midi-input-current-replacement-pattern pattern)
    (apply 'message
	   (concat "Replacement pattern: " midi-input-current-replacement-pattern)
	   (make-list nmatches "a"))))


(defun midi-input-transpose-region-command (beg end fifths)
"Transpose region tis many fifths along the circle.
Only usable in relative mode "
 (interactive
      (let ((string (read-string "How many steps along circle of fifths? (positive or negative number) ")))
             (list (region-beginning) (region-end) (string-to-number string))))
 (let*( (region
	 (buffer-substring-no-properties beg end))
	(bare-notename-regexp
	 (format "\\(%s\\)\\(%s\\)\\(%s\\)"
		 midi-input-before-notename-regexp
		 midi-input-notename-regexp
		 midi-input-after-notename-regexp))
	(transposed-region
	 (car (midi-input-replace-all-matches bare-notename-regexp region  (lambda (note) (midi-input-transpose fifths note)) 2))))

   (save-excursion (goto-char beg)
		   (delete-region beg end)
		   (insert transposed-region))))





(defun midi-input-replace-all-matches (regex string1 fun &optional subexp)

  "Replace all matches for regex (or for its <subexp>'th parenthesised 
subexpression) in string1 by (fun match) and return a cons
pair of the result and the number of matches found
Any matches for regex in string2 will not themselves be matched"
  
  (let* ((word string1)
	 (pos 0)
	 (count 0)
	 (len (length word))
	 (oldlen len)
	 (matchpos 0)
	 (subexno
	  (if (null subexp) 0 subexp)))
    (while (setq matchpos (string-match regex word pos))
      ;; (message "word: %s matchpos: %d pos: %d count: %d len: %d oldlen: %d"  word matchpos pos count len oldlen)
      (setq pos (match-end subexno))
      (setq word (replace-match  (apply fun (match-string subexno word) nil)  nil t word subexp)) ;; (apply fun subexp nil)
      (setq len (length word))
      (setq pos (+ pos (- len oldlen))) ; if word has grown or shrunken, update pos accordingly
      (setq count (1+ count))
      (setq oldlen len))  
    (cons word count)))

;;; (apply '(lambda (x) (+ 3 x)) 6 nil)

;;; TRYME (midi-input-replace-all-matches "\\( \\)\\(a\\(is\\)?\\('+\\|,+\\)?\\)\\( \\)" " a ais, a'''' a', "  '(lambda (s) "bla")  2)
;;; TRYME (midi-input-replace-all-matches midi-input-notename-regexp "bes, is  fis f"  (lambda (s) (midi-input-transpose 3 s) )  0)

(defun midi-input-count-matches (regex string)
  (let*(  (count 0)
	  (pos 0))
    (while (string-match regex string pos)
      (setq pos (match-end 0))
      (setq count (1+ count)))
    count))

;;; TRYME (midi-input-count-matches "[aeiou]+" "bewaarplaatsen")

(defun midi-input-replace-pattern (&optional pattern)
  (interactive)
  (if (not pattern)
      (setq pattern midi-input-current-replacement-pattern))
  (save-excursion
    (catch 'badnotes
      (let*((len
	     (midi-input-count-matches "%s" pattern))  
	    (chunks
	     (midi-input-splice  (make-list len "[^ \\t]+")  "[ \\t]+"))
	    (regexp
	     (apply 'concat "\\<" chunks)) ; make the first chunk match a whole notename (e.g. bes and not just es)
	    (from
	     (condition-case nil
		 (posix-search-backward regexp (save-excursion (progn (beginning-of-line) (point))))
	       (error (throw 'badnotes (message "Not enough notes before point")))))		
	    (text
	     (match-string 0))
	    (notes
	     (save-match-data (split-string text)))
	    (dummy
	     (message (apply 'concat (midi-input-splice notes "/"))))
	    (testfunc
	     '(lambda (maybe-a-note)
		(if (not (save-match-data (string-match midi-input-notename-regexp maybe-a-note)))
		    (throw 'badnotes (message "Cannot find notename in %s" maybe-a-note)))))
	    (test-them-all
	     (mapcar testfunc notes))
	    (new-text
	     (apply 'format pattern notes)))
	(replace-match new-text t t)))))




(defun midi-input-toggle-help ()
  (interactive)
  (let* ((bufname 
	  "Midi input help")
	 (buf 
	  (get-buffer bufname))
	 )
    
    (if buf (kill-buffer buf)
      (setq buf (generate-new-buffer bufname))
      (set-buffer buf)      
      (let ((helptext "")
	     sorted-shiftkeys)
	(setq sorted-shiftkeys
	      (copy-list (append 
			  midi-input-user-defined-shifted-keys 
			  midi-input-shifted-keys)))
	(setq sorted-shiftkeys 
	      (sort sorted-shiftkeys '(lambda (x y) (< (car x) (car y)))))
	(while (not (null sorted-shiftkeys))
	  (let*((key
		 (car sorted-shiftkeys))
		(keyno 
		 (elt key 0))
		(keyname
		 (midi-input-noteno-to-notename keyno t))
		(action 
		 (elt key 1))
		(comment 
		 (elt key 2))
		(actionstring 
		 (if (stringp action)
		     (format "%S" action)
		   (format "%S" action)))
		(commentstring
		 (if (> (length comment) 0) comment nil)))
	    (insert (format "%-5s: %s\n" keyname 
	      (if commentstring commentstring actionstring)))	
	    (setq sorted-shiftkeys (cdr sorted-shiftkeys)))))     
      (switch-to-buffer buf))))

(midi-input-initialise)
	
;;; Local variables:
;;; mode: outline-minor
;;; End:

      